from pydantic import BaseModel


class Settings(BaseModel):
    appname: str
    timeout: float = 0.1
    port: int = 9999
    volume: int = 100
    rate: int = 4
    chunk: int = 4096
    speak_timeout = 10_000



# Standard lib
from threading import Thread
from typing import Iterator
from io import BytesIO
import audioop
import queue

# Third party
import pyaudio

# Local
from willspeak import settings, inactive_flag


class SpeakerBase(Thread):
    """
    Base class for all audio speakers.
    """

    def __init__(self, *args, **kwargs):
        super(SpeakerBase, self).__init__(*args, **kwargs)
        self.queue = queue.Queue()

    def stop(self):
        """
        Stop current playback and clear any unprocessed chunks.
        """
        # Clear the queue of audio chunks
        with self.queue.mutex:
            self.queue.queue.clear()

    def play(self, audio_data: bytes):
        """
        Split audio data into chunks and add to processing queue.
        """
        audio_obj = BytesIO(audio_data)
        while dataframe := audio_obj.read(settings.chunk):
            self.queue.put(dataframe)

    def play_chunk(self, audio_chunk: bytes):
        """
        Take an audio chunk and add to processing queue.
        """
        self.queue.put(audio_chunk)

    def play_iterator(self, audio_stream: Iterator[bytes]):
        """
        Take an iterator of audio chunks and add them to processing queue.
        """
        for audio_chunk in audio_stream:
            self.queue.put(audio_chunk)

    def close(self):
        """Close any opened audio streams."""
        raise NotImplementedError()

    def _play_stream(self, audio_data: bytes):
        """Internal method that will play the given audio stream."""
        raise NotImplementedError()

    def run(self) -> None:
        """Continuously wait for and process any audio data."""
        try:
            while not inactive_flag.is_set():
                try:
                    audio_data = self.queue.get(timeout=settings.timeout)
                except queue.Empty:
                    continue

                try:
                    self._play_stream(audio_data)
                finally:
                    self.queue.task_done()
        finally:
            self.close()


class WaveSpeaker(SpeakerBase):
    """Speaker thread to handle processing of wave audio streams."""
    def __init__(self, *args, **kwargs):
        super(WaveSpeaker, self).__init__(*args, **kwargs)
        self._pyaudio = pyaudio.PyAudio()
        self._stream = self._create_stream()

    def _create_stream(self) -> pyaudio.Stream:
        """Create a pyaudio stream in stereo."""
        return self._pyaudio.open(format=pyaudio.paInt16, channels=2, rate=48000, output=True)

    def _play_stream(self, audio_data: bytes):
        # Convert audio chunk from mono to stereo before playing
        audio_data = audioop.tostereo(audio_data, 2, 1, 1)
        self._stream.write(audio_data)

    def close(self):
        # Close the audio stream
        self.stop()
        self._stream.close()
        self._pyaudio.terminate()



# Standard lib
import threading

# Local
from .conf import Settings

# Event flag to trigger graceful shutdown
inactive_flag = threading.Event()

# Project settings
settings = Settings(
    appname="WillSpeak"
)



# Standard lib
from dataclasses import dataclass, field, fields
import typing


@dataclass(eq=False, frozen=True)
class Voice(object):
	"""
	Voice dataclass.

	:ivar id: ID of the voice.
	:ivar name: The voice name.
	:ivar lang: The language of the voice, including the region, (United States) or (Great Britain).
	:ivar gender: The gender, male or female.
	:ivar age: The age group of the vocie. e.g. Adult, child.
	"""

	id: int | str
	name: str
	lang: str = field(kw_only=True, default="")
	gender: str = field(kw_only=True, default="")
	age: str = field(kw_only=True, default="")
	token: typing.Any = field(kw_only=True, default=None, repr=False)

	def asdict_shallow(self) -> dict:
		"""Converts the dataclass obj to a dict using a shallow copy only."""
		# We do not want the token variable to be return, token is internal
		return dict((_field.name, getattr(self, _field.name)) for _field in fields(self) if _field.name != "token")



# Standard lib
import functools
import typing

# Local
from willspeak import inactive_flag


def graceful_exception(func):
    """
    Decorator function to handle exceptions gracefully.
    And signal any threads to end.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwargs) -> int:
        try:
            ret = func(*args, **kwargs)
        except KeyboardInterrupt:
            return 143
        else:
            # Return 0 if no return code was returned
            return ret if ret is not None else 0
        finally:
            inactive_flag.set()
    return wrapper


def ensure_int_range(min_value: int, max_value: int) -> typing.Callable[[str], int]:
    """
    Function that converts a string into an integer while ensuring the value is within the given range.

    :param min_value: The minimum allowed value.
    :param max_value: The maximum allowed value.
    """
    def wrapper(string: str) -> int:
        value = int(string)
        if min_value <= value <= max_value:
            return value
        else:
            raise ValueError(f"'{value}' is outside range, must be anywhere from {min_value} and {max_value}")

    return wrapper



"""
This module contains filter functions that should filter out and/or manipulate the text it's processing.

There are 3 different types of filters.
    * Decider: Function that checks if given text should be processed or not. Returning None if not.
    * Filter: Simple function that takes a given text and filters out any unwanted text before returning it.
    * Processor: This function takes a given text and splits it up into more manageable parts. Faster when
                 dealing with larger chunks of text.
"""

# Standard lib
from typing import Callable
import re

T_FILTER_RTN = str | list[str] | None
T_FILTER_CALL = Callable[[str], T_FILTER_RTN]
text_filters: list[T_FILTER_CALL] = []
regex_cache = {}


def process_text(text: str) -> list[str]:
    """Run all filters & processers on given text string."""
    text_list = [text]
    for func in text_filters:
        new_list = []
        for text in text_list:
            ret = func(text)
            if isinstance(text, list):
                new_list.extend(ret)
            elif text:
                new_list.append(ret)

        text_list = new_list

    return text_list


def re_cache(regex: str):
    """Takes regex string, compile it, and cache it for later use."""
    if regex in regex_cache:
        return regex_cache[regex]
    else:
        compiled = re.compile(regex)
        regex_cache[regex] = compiled
        return compiled


def text_filter(func) -> T_FILTER_CALL:
    """
    Function decorator to register a text filter.

    A text filter takes a string and strips out unwanted text patterns.
    """
    def wrapper(text):
        try:
            return func(text)
        except Exception as e:
            print("ERROR:", e)
            return text

    text_filters.append(wrapper)
    return wrapper


# Deciders
################################################################################

@text_filter
def ignore_only_url(text: str) -> str | None:
    """If the given text only contains an url then ignore it."""
    re_filter = re_cache(r"^(http|ftp|https):\/\/([\w_-]+(?:(?:\.[\w_-]+)+))([\w.,@?^=%&:\/~+#-]*[\w@?^=%&\/~+#-])$")
    # Returns text only if match is not found
    if not re.match(re_filter, text):
        return text


# Filters
################################################################################

@text_filter
def filter_read_more_at(text: str) -> str:
    """
    Strip out any 'Read more at.' text that a lot
    of sites like to add when any text is copied.
    """
    re_filter = re_cache(r"([Rr]ead more\s*[at]*:*\s+http\S+|[Ss]ee more at:*\s+http\S+)")
    return re.sub(re_filter, "", text).strip()


# Processors
################################################################################



# Standard lib
import asyncio

# Third party
from aiohttp import web

# Allow for route decorators
routes = web.RouteTableDef()


# Functions
######################################################################

def create_full_url(request: web.Request, path: str) -> str:
    """Create a full absolute url using request to build the url."""
    return f"{request.scheme}://{request.host}{path}"


# Views
######################################################################

@routes.get("/")
async def api_root(request: web.Request):
    """
    This is the api root. It should list the urls to all the api endpoints.

    Endpoints:
        * /speak: Convert text to speach and respond with a raw wav file.
        * /settings: Get/set TTS engine parameters.
        * /voices: Get list of all available voices.
    """
    data = dict(
        speak=create_full_url(request, "/speak"),
        stop=create_full_url(request, "/stop"),
        settings=create_full_url(request, "/settings"),
        voices=create_full_url(request, "/voices"),
    )
    return web.json_response(data)


@routes.get("/voices")
async def get_voices(request: web.Request):
    """
    Return list of all available voices.
    """
    engine = request.app["tts_engine"]
    data = []
    for voice in engine.get_voices():
        voice_dict = voice.asdict_shallow()
        voice_dict["url"] = create_full_url(request, f"/voices/{voice_dict['id']}")
        data.append(voice_dict)
    return web.json_response(data)


@routes.get("/voices/{id}")
async def get_voice(request: web.Request):
    """
    Return details about slected voice.
    """
    engine = request.app["tts_engine"]
    voice_id = request.match_info["id"]

    for voice in engine.get_voices():
        # Do the check as strings, as the voice ID can be a string
        if str(voice.id) == str(voice_id):
            voice_dict = voice.asdict_shallow()
            voice_dict["url"] = create_full_url(request, f"/voices/{voice_dict['id']}")
            return web.json_response(voice_dict)

    message = f"No voice found matching ID: {voice_id}"
    return web.json_response({"error": message}, status=404)


@routes.view("/settings")
class Settings(web.View):
    async def get(self):
        """
        Return the current settings values.

        Settings:
            * rate: The current speaking rate.
            * volume: The current speaking volume.
            * voice: The currently selected voice.
        """
        engine = self.request.app["tts_engine"]
        voice_dict = engine.voice.asdict_shallow()
        voice_dict["url"] = create_full_url(self.request, f"/voices/{voice_dict['id']}")

        data = dict(
            rate=engine.rate,
            volume=engine.volume,
            voice=voice_dict,
        )

        return web.json_response(data)

    async def put(self):
        """
        Change any or all of the tts settings.
        """
        engine = self.request.app["tts_engine"]
        req_data = await self.request.json()

        if "rate" in req_data:
            engine.rate = req_data["rate"]
        if "volume" in req_data:
            engine.volume = req_data["volume"]
        if "voice" in req_data:
            voice_id = req_data["voice"]
            try:
                engine.voice = voice_id
            except ValueError as e:
                return web.json_response({"error": str(e)}, status=400)

        # Return the full settings response with changes
        return await self.get()


@routes.view("/speak")
class Speak(web.View):
    """
    Convert any given text into speach.

    The text can be given by using url params, or by using the request body.
    The url param to use is called 'text', the value needs to be url encoded.
    When sending the text using the body it can just be a plain text post request.
    The response will be a raw wave file.

    Support for other response types are planned.
    """
    EOF = object()

    async def get(self):
        """Handle text from url params."""
        if "text" not in self.request.query:
            message = "Missing required url parameter: text"
            return web.json_response({"error": message}, status=400)

        texts = self.request.query.getall("text")
        text = " ".join(texts)
        return await self.speak(text)

    async def post(self):
        """Handel text from post request."""
        if not self.request.can_read_body:
            message = "Missing required text body"
            return web.json_response({"error": message}, status=400)

        text = await self.request.text()
        return await self.speak(text)

    async def speak(self, text: str):
        """Convert text into speach."""
        loop = asyncio.get_running_loop()
        resp = web.StreamResponse()
        await resp.prepare(self.request)

        def callback(audio_chunk: bytes):
            # Be careful here, this code is synchronous, no async here
            coroutine = resp.write(audio_chunk)
            asyncio.run_coroutine_threadsafe(coroutine, loop).result()

        # Call TTS engine in another thread as the call is a blocking call
        engine = self.request.app["tts_engine"]
        stream = engine.create_memory_stream()
        await loop.run_in_executor(None, engine.raw_speak, stream, callback, text)

        # Finalize response
        await resp.write_eof()
        return resp


@routes.get("/stop")
async def stop_speak(request: web.Request):
    engine = request.app["tts_engine"]
    engine.stop()
    return web.json_response(status=204)


# Async Setup
######################################################################

@web.middleware
async def http_404_middleware(request, handler):
    """
    Middleware that converts a 404 page
    not found error into a json response.
    """
    try:
        response = await handler(request)
        if response.status != 404:
            return response
        message = response.reason
    except web.HTTPException as ex:
        if ex.status != 404:
            raise
        message = ex.reason
    return web.json_response({"error": message})


async def setup_tt_engine(app):
    """Initialize TTS engine."""
    from willspeak.tts.sapi5.local import TTSClient
    engine = TTSClient(server_mode=True)
    app["tts_engine"] = engine


def start(addr: str, port: int):
    """Start the async http sapi5 api server."""
    middlewares = [http_404_middleware]
    app = web.Application(middlewares=middlewares)
    app.add_routes(routes)
    app.on_startup.append(setup_tt_engine)
    web.run_app(app, host=addr, port=port)



# Local
from .monitors import ClipboardMonitor, ProcessMonitor
from .utils import graceful_exception
from .audio import WaveSpeaker


@graceful_exception
def stop(_):
    """Send a stop command to running process."""
    process_pipe = ProcessMonitor()
    process_pipe.send_stop()
    return 0


@graceful_exception
def server(args):
    """Set up the TTS engine in server mode."""
    from .tts.sapi5 import server as sapiserver
    sapiserver.start(args.bind_addr, args.bind_port)


@graceful_exception
def client(args):
    """Set up the TTS engine in client mode. Requires a running tts server."""
    from willspeak.tts.sapi5.client import TTSClient
    return speaking_mode(TTSClient, args, addr=args.addr, port=args.port)


@graceful_exception
def local(args):
    """Set up the TTS engine in local mode."""
    from willspeak.tts.sapi5.local import TTSClient
    return speaking_mode(TTSClient, args)


def speaking_mode(tts_engine, args, **init_args):
    """Run the tts engine, monitor clipboard and speak any text."""
    # Welcome the user
    tts = tts_engine(**init_args)
    tts.volume = args.volume
    tts.rate = args.rate

    # Welcome user
    tts.speak("Welcome to the new will speak!")

    # Allow for process communication to control current state
    process_pipe = ProcessMonitor()
    process_pipe.start_server(stop=tts.stop)

    # Monitor clipboard for text to speak
    monitor = ClipboardMonitor()
    monitor.wait_for_text(tts.speak)
    return 0



# Standard lib
from typing import Callable
import hashlib

# Third party
# noinspection PyUnresolvedReferences
from comtypes.client import CreateObject

# Local
from ..voice import Voice
from ...audio import WaveSpeaker
from ... import settings

# Stream Formats
SAFT22kHz16BitMono = 22
SAFT22kHz16BitStereo = 23
SAFT24kHz16BitMono = 26
SAFT24kHz16BitStereo = 27
SAFT32kHz16BitMono = 30
SAFT32kHz16BitStereo = 31
SAFT44kHz16BitMono = 34
SAFT44kHz16BitStereo = 35
SAFT48kHz16BitMono = 38
SAFT48kHz16BitStereo = 39

# Speak Flags
SVSFDefault = 0
SVSFlagsAsync = 1
SVSFPurgeBeforeSpeak = 2
SVSFIsFilename = 4
SVSFIsXML = 8
SVSFIsNotXML = 16
SVSFPersistXML = 32


def create_voice_obj(cvoice) -> Voice:
	"""Takes a C SpObjectToken voice object converting to a python data object."""
	# The description contains both the name and lang
	# The real lang attribute is a number (strange)
	full_name = cvoice.GetDescription()
	name, lang = full_name.split(" - ", 1)
	name = name.replace("Desktop", "").replace("Mobile", "").strip()
	lang = lang.strip()

	# We create ID instead of using the available one
	# The available one is a mix of version and name with some other crap
	voice_id = int(str(int(hashlib.sha1(full_name.encode("utf8")).hexdigest(), 16))[:8])

	return Voice(
		id=voice_id,
		name=name,
		lang=lang,
		age=cvoice.GetAttribute("Age"),
		gender=cvoice.GetAttribute("Gender"),
		token=cvoice,
	)


class TTSClient(object):
	"""
	TTS driver module to control microsoft SAPI TTS Engine.
	"""

	speaker_class = WaveSpeaker

	def __init__(self, server_mode=False):
		self._engine = CreateObject("SAPI.SPVoice")
		if not server_mode:
			self._speaker = self.speaker_class()
			self._speaker.start()
		else:
			self._speaker = None

	# noinspection PyMethodMayBeStatic
	def create_memory_stream(self):
		"""
		Create SAPI in memory audio stream.
		"""
		stream = CreateObject("SAPI.SpMemoryStream")
		# Mono keeps the transfer size down
		# Audio stream can always be converted to stereo after the fact
		stream.Format.Type = SAFT48kHz16BitMono
		return stream

	def get_voices(self) -> list[Voice]:
		"""
		Return list of avalable voices as a Voice object.
		"""
		return [create_voice_obj(attr) for attr in self._engine.GetVoices()]

	@property
	def voice(self) -> Voice:
		"""
		Return the current selected voice as a Voice object.
		"""
		voice = self._engine.Voice
		return create_voice_obj(voice)

	@voice.setter
	def voice(self, value: Voice | int | str):
		"""
		Change voice to selected voice.

		Value can be a voice object, voice id (int) or voice name (str).
		"""
		if isinstance(value, Voice):
			self._engine.Voice = value.token
			return

		# Search for voice using given id or name
		for voice in self.get_voices():
			# Do the check as strings, as the voice ID can be a string
			if str(voice.id) == str(value) or voice.name == value:
				self._engine.Voice = voice.token
				break
		else:
			message = f"No voice found matching ID: {value}"
			raise ValueError(message)

	@property
	def rate(self) -> int:
		"""
		Return the current speech rate.
		"""
		return self._engine.Rate

	@rate.setter
	def rate(self, rate: int):
		"""
		Change the speach rate of the voice.

		value : integer --- Rate to set the voice to.

		Value must be between -10 to +10.
		"""
		if rate < -10 or rate > 10:
			raise ValueError("invalid rate %s" % rate)
		else:
			self._engine.Rate = rate

	@property
	def volume(self) -> int:
		"""
		Return the current volume level.
		"""
		return self._engine.Volume

	@volume.setter
	def volume(self, value: int):
		"""
		Chagne the volume of the voice.

		value : int --- level of the volume to change to.
		"""
		if value < 0 or value > 100:
			raise ValueError("invalid volume %s" % value)
		else:
			self._engine.Volume = value

	def speak(self, text: str):
		"""
		Send text to tts engine and speak.
		"""
		callback = self._speaker.play_chunk
		stream = self.create_memory_stream()
		self.raw_speak(stream, callback, text)

	def raw_speak(self, stream, callback: Callable[[bytes], None], text: str) -> bool:
		"""
		Call the speach engin and send all audio to the given stream.

		:returns bool: Returns True if command completed, False if command timed out.
		"""
		self._engine.AudioOutputStream = stream
		self._engine.Speak(text, SVSFlagsAsync)

		# Wait 10 seconds for the conversion to complete
		# Any longer than that and there must be an issue
		# WaitUntilDone returns False when it times out
		completed = not self._engine.WaitUntilDone(settings.speak_timeout)
		if completed:
			self.stop()
			self._engine.Speak("Error: Speak Request timed out")

		stream.Seek(0)
		while True:
			audio_chunk, data_size = stream.read(settings.chunk)
			if data_size:
				callback(bytes(audio_chunk))
			else:
				break

		return completed

	def stop(self):
		"""
		Stop current speak request.
		"""
		self._engine.Speak("", SVSFPurgeBeforeSpeak)
		if self._speaker is not None:
			self._speaker.stop()

	"""
	# May not need this, Just here for reference
	def speak_to_file(self, text, filepath):
		file_stream = CreateObject("SAPI.SPFileStream")
		file_stream.Format.Type = 39
		file_stream.Open(filepath, 3)
		self._engine.AudioOutputStream = file_stream
		self._engine.Speak(text)
		file_stream.Close()
		self._engine.AudioOutputStream = self._stream
	"""



# Standard lib
import argparse
import sys

# Local
from willspeak import settings, main
from willspeak.utils import graceful_exception, ensure_int_range


@graceful_exception
def entrypoint() -> int:
    # We use argparse commands to select the function to call
    args = parser.parse_args(sys.argv[1:] or ["-h"])
    return args.func(args)


def tts_client_settings(sub_parsser):
    """
    Add volume & rate to arg parser.
    """
    # Add volume setting
    sub_parsser.add_argument(
        "--volume",
        help="Set the tts speaking volume. Value can be anyting from 0 to 100",
        default=settings.volume,
        type=ensure_int_range(0, 100),
    )
    # Add speaking rate setting
    sub_parsser.add_argument(
        "--rate",
        help="Set the tts speaking rate. Value can be anyting from 1 to 10",
        default=settings.rate,
        type=ensure_int_range(-10, 10),
    )


# CLI Arguments
parser = argparse.ArgumentParser(prog=settings.appname)
subcommands = parser.add_subparsers(title="commands", required=True, dest="cmd")
parser.add_argument(
    "-e", "--engine",
    default="sapi5",
    choices=["sapi5"],
    help="Text to speach engine to use. Default('sapi5')"
)

# Server Command
sub = subcommands.add_parser("server", help=f"Run {settings.appname} in server mode")
sub.set_defaults(func=main.server)
sub.add_argument(
    "--bind_addr",
    help=f"The IP interface to bind to, default is 0.0.0.0 (all addresses).",
    metavar="",
    default="0.0.0.0",
)
sub.add_argument(
    "--bind_port",
    help="The port number to listen on.",
    metavar="",
    default=settings.port,
    type=int,
)

# Client Command
sub = subcommands.add_parser("client", help=f"Run {settings.appname} in server mode")
sub.set_defaults(func=main.client)
tts_client_settings(sub)
sub.add_argument(
    "--addr",
    help=f"The IP/Hostname of the {settings.appname} server.",
    metavar="",
    required=True,
)
sub.add_argument(
    "--port",
    help="The port number to connect to on the server.",
    metavar="",
    default=settings.port,
    type=int,
)

# Local Command
sub = subcommands.add_parser("local", help=f"Run {settings.appname} in local mode")
sub.set_defaults(func=main.local)
tts_client_settings(sub)

# Stop Command
sub = subcommands.add_parser("stop", help=f"Stop current speach and clear any queue")
sub.set_defaults(func=main.stop)


# This is only used for manual testing
if __name__ == "__main__":
    exit_code = entrypoint()
    sys.exit(exit_code)



"""
This module monitors the clipboard for changes and add any copied
text to a processing queue.
"""

# Standard lib
from multiprocessing.connection import Listener, Client
from threading import Thread
from typing import Callable
import queue

# Third party
import pyperclip

# Local
from willspeak import inactive_flag, settings, filters


class ClipboardMonitor(Thread):
    """Monitors the clipboard for changes and adds to processing queue."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._process_queue = queue.Queue()

    # noinspection PyUnusedLocal
    def run(self) -> None:
        """
        This function call blocks until a new text string exists on the
        clipboard that is different from the text that was there when the function
        was first called.
        """
        # We ignore the first clipboard text on startup
        prev_text = self.get_clipboard()
        while not inactive_flag.wait(settings.timeout):
            text = self.get_clipboard()
            # Ensure that we don't have the same text again
            if text and text != prev_text:
                prev_text = text
                self._process_queue.put(text)

    # noinspection PyMethodMayBeStatic
    def get_clipboard(self) -> str:
        """Return text from clipboard."""
        return pyperclip.paste().strip()

    def wait_for_text(self, callback: Callable[[str], None]):
        """
        Continuously monitor the queue for text to convert to speach.
        Filtering the text as it runs.
        """
        # This will start the background clipboard monitor
        self.start()

        # This will wait for available text
        while not inactive_flag.is_set():
            try:
                text = self._process_queue.get(timeout=settings.timeout)
            except queue.Empty:
                continue

            try:
                # Process text before speaking
                for text in filters.process_text(text):
                    callback(text)
            finally:
                self._process_queue.task_done()


class ProcessMonitor(Thread):
    LISTEN_PORT = 33950

    def __init__(self, *args, **kwargs):
        super().__init__(*args, daemon=True, **kwargs)
        self._commands = None

    def run(self) -> None:
        listener = Listener(("127.0.0.1", self.LISTEN_PORT))

        while not inactive_flag.is_set():
            conn = listener.accept()
            command = conn.recv()
            if command in self._commands:
                func = self._commands[command]
                func()

            conn.close()
        listener.close()

    def start_server(self, **commands):
        """Start the process listener server to check for process comamnds."""
        self._commands = commands
        self.start()

    def send_stop(self):
        """Send the stop signal to master process."""
        try:
            conn = Client(("127.0.0.1", self.LISTEN_PORT))
        except ConnectionRefusedError:
            pass
        else:
            conn.send("stop")
            conn.close()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['willspeak', 'willspeak.tts', 'willspeak.tts.sapi5']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.3,<4.0.0',
 'pyaudio>=0.2.12,<0.3.0',
 'pydantic>=1.10.2,<2.0.0',
 'pyperclip>=1.8.2,<2.0.0',
 'requests>=2.28.1,<3.0.0']

extras_require = \
{':sys_platform == "win32"': ['comtypes>=1.1.14,<2.0.0']}

entry_points = \
{'console_scripts': ['willspeak = willspeak.cli:entrypoint']}

setup_kwargs = {
    'name': 'willspeak',
    'version': '0.3.0',
    'description': 'Python Text to Speach using Microsoft Sapi5 with a server/client model',
    'long_description': '# WillSpeak - Work in Progress\nPython Text to Speach using Microsoft Sapi5 with a server/client model.\n\n# Progress update\nThe core functionality is now working, and is ready for testing.\nSome cleanup is still required, but it works.\nOnly supports SAPI5 for now. More to come in the future.\n\n# Info\nI created this project as a way to have good TTS on linux, because TTS on linux at the moment is dreadful.\nFor a long time I wanted to switch to linux, but I needed a good linux TTS software but could not find one.\nSo I decided to create this project to interface with the windows SAPI5 TTS engine.\n\nHow it works is by running this software in server mode on a Windows machine. \nThen configure the linux client to communicate with that Windows TTS server.\nThe client will monitor for text that was copied to the clipboard and converts the text into speech.\n\n# Usage\nThis software has 2 different operational modes, "Local" & "Server/Client". If the TTS engine that you have selected \nworks natively on your operating system, Then you can use Local mode. e.g. SAPI5 is native to windows, so you can use\nLocal mode on Windows when using SAPI5. You should use Server/Client if you want to use SAPI5 on linux.\n\nRun locally on Windows\n```shell\nwillspeak local\n```\n\nTo run in server mode do.\n```shell\nwillspeak server\n```\n\nAnd on the client machine run. "--addr" is the address of the server running the server component.\n```shell\n# 192.168.1.60 is just an example\nwillspeak client --addr=192.168.1.60\n```\n\nThere is one last command that is used to stop any current speech.\n```shell\nwillspeak stop\n```\n\n# TODO\n* Use a string library to analyze and filter the text before converting.\n* Setup prometheus metrics to track usage. This is useful if you wish to use a paid for TTS Service.\n* Add support for other text to speech engines like eSpeak.\n* Add support for running the server component as a Windows service.\n\n# Links\nhttps://winaero.com/unlock-extra-voices-windows-10/\n\n## Version\n0.3.0\n',
    'author': 'willforde',
    'author_email': 'willforde@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/willforde/willspeek.git',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



# Third party
import requests

# Local
from ..voice import Voice
from ...audio import WaveSpeaker


class TTSClient(object):
	"""
	TTS driver module to control microsoft SAPI TTS Engine.
	"""

	# Methods
	GET = "GET"
	PUT = "PUT"
	POST = "POST"

	# Endpoints
	EPT_SETTINGS = "/settings"
	EPT_VOICES = "/voices"
	EPT_SPEAK = "/speak"
	EPT_STOP = "/stop"

	speaker_class = WaveSpeaker

	def __init__(self, addr: str, port: int):
		self.base_url = f"http://{addr}:{port}"
		self.speaker = self.speaker_class()
		self.speaker.start()

		# Requests
		self.session = requests.Session()

	def _make_json_request(self, method: str, url: str, **kwargs):
		url = self.base_url + url
		resp = self.session.request(method, url, **kwargs)
		resp.raise_for_status()
		resp_datta = resp.json()
		resp.close()
		return resp_datta

	def get_voices(self) -> list[Voice]:
		"""
		Return list of avalable voices as a Voice object.
		"""
		voices = []
		for voice in self._make_json_request(self.GET, self.EPT_VOICES):
			del voice["url"]
			voice_obj = Voice(**voice)
			voices.append(voice_obj)

		return voices

	@property
	def voice(self) -> Voice:
		"""
		Return the current selected voice as a Voice object.
		"""
		resp = self._make_json_request(self.GET, self.EPT_SETTINGS)
		del resp["voice"]["url"]
		return Voice(**resp["voice"])

	@voice.setter
	def voice(self, voice: Voice | int | str):
		"""
		Change voice to selected voice.

		Value can be a voice object, voice id (int) or voice name (str).
		"""
		if isinstance(voice, Voice):
			voice = voice.id

		req_data = {"voice": voice}
		self._make_json_request(self.PUT, self.EPT_SETTINGS, json=req_data)

	@property
	def rate(self) -> int:
		"""
		Return the current speech rate.
		"""
		resp = self._make_json_request(self.GET, self.EPT_SETTINGS)
		return resp["rate"]

	@rate.setter
	def rate(self, rate: int):
		"""
		Change the speach rate of the voice.

		value : integer --- Rate to set the voice to.

		Value must be between -10 to +10.
		"""
		req_data = {"rate": rate}
		self._make_json_request(self.PUT, self.EPT_SETTINGS, json=req_data)

	@property
	def volume(self) -> int:
		"""
		Return the current volume level.
		"""
		resp = self._make_json_request(self.GET, self.EPT_SETTINGS)
		return resp["volume"]

	@volume.setter
	def volume(self, volume: int):
		"""
		Chagne the volume of the voice.

		value : int --- level of the volume to change to.
		"""
		req_data = {"volume": volume}
		self._make_json_request(self.PUT, self.EPT_SETTINGS, json=req_data)

	def speak(self, text):
		"""
		Send text to tts engine and speak.
		"""
		raw_text = text.encode("utf8")
		url = self.base_url + self.EPT_SPEAK
		headers = {"content-type": "text/plain; charset=UTF-8", "content-length": str(len(raw_text))}
		resp = self.session.request(self.POST, url, data=raw_text, headers=headers, stream=True)
		resp.raise_for_status()

		# Stream the response into the speaker
		iterator = resp.iter_content(4096)
		self.speaker.play_iterator(iterator)
		resp.close()

	def stop(self):
		"""
		Stop current speak request.
		"""
		self._make_json_request(self.GET, self.EPT_STOP)

	def close(self):
		self.session.close()
